/*
  ggHiggs code based on

    R. D. Ball, M. Bonvini, S. Forte, S. Marzani, G. Ridolfi
    arXiv:1303.3590

  please cite it if you use it
*/

#include <iomanip>
#include <vector>
#include <algorithm>
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <stdio.h>
#include <cmath>
#include <ctime>
#include <cstdlib>
#include <sys/stat.h> 
#include <sys/time.h> 
#include <getopt.h> 

#include "LHAPDF/LHAPDF.h"  // interface with LHAPDF

#include "./src/ggHiggs.hh"
#include "./src/finite-mt_small-x_coeffs.hh"



using namespace std;



/*
  If you like, you can customize your alpha_s routine.
  However, consistency with PDF sets suggests the use of the LHAPDF alphas routine.
*/
double alphas(double mu) {
  return LHAPDF::alphasPDF(mu);
  //return alphas_ihixs(mu, order, LHAPDF::alphasPDF(mZ), mZ); // taken from ihixs code
}





void print_usage() {
  cout << endl
       << "Usage:" << endl
       << " -t  --finite-mt        use finite top mass (default: infinite-mtop approximation)" << endl
       << " -p  --pdf-set <set>    specify PDFs being used (default \"NNPDF23_nnlo_as_0118\")" << endl
       << " -m  --mass <mass>      set Higgs mass [GeV]" << endl
       << " -E  --Ecm <energy>     set collider energy [GeV]" << endl
       << " -f  --fact <murat>     set muF/mu  ratio" << endl
       << " -r  --ren <murat>      set muR/muF ratio" << endl
       << endl;
  exit(0);
  return;
}






const string sorder[4] = {"LO", "NLO", "NNLO", "NNNLO"};
void read_arguments(int argc, char* argv[], double &mH, double &sqrts, string &set) {
  const char* const short_options = "htp:f:r:m:E:";
  const struct option long_options[] = { { "help", 0, NULL, 'h' },
					 { "finite-mt", 0, NULL, 't' },
					 { "pdf-set", 1, NULL, 'p' },
					 { "fact", 1, NULL, 'f' },
					 { "ren", 1, NULL, 'r' },
					 { "mass", 1, NULL, 'm' },
					 { "Ecm", 1, NULL, 'E' },
					 { NULL, 0, NULL, 0 } };
  int next_option;
  ostringstream sset;
  do {
    next_option = getopt_long (argc, argv, short_options, long_options, NULL);
    switch (next_option) {
    case 'h':
      print_usage();
    case 't':
      finite_mt = true;
      break;
    case 'p':
      sset << optarg;
      set = sset.str();
      break;
    case 'f':
      _muF_mu_ratio = strtod(optarg, NULL);
      break;
    case 'r':
      _muR_muF_ratio = strtod(optarg, NULL);
      break;
    case 'm':
      mH = strtod(optarg, NULL);
      break;
    case 'E':
      sqrts = strtod(optarg, NULL);
      break;
    case '?':
      print_usage();
    case -1: break;
    default: abort();
    }
  }
  while (next_option != -1);
  //
  // messages
  if(!APimprExponent) cout << endl << "  !!! WARNING !!! " << endl << " AP improvement not exponentiated!" << endl << endl;
  if(!finite_mt) cout << ">>> Infinite m_t limit <<<" << endl;
  else           cout << ">>> Finite m_t !! <<<" << endl;
  //
  cout << ">>> Order: " << sorder[order] << endl;
  //
  cout << ">>> PDFset: " << set << endl;
  //
  cout << "Renormalization scale muR = " << _muR_muF_ratio*_muF_mu_ratio << " * mH" << endl;
  cout << "Factorization   scale muF = " << _muF_mu_ratio << " * mH" << endl;
  return;
}







int main (int argc, char* argv[]) {


  /*************  input parameters  **************/

  // masses (GeV)
  mtop = 172.5; // 172.5 is consistent with small-x coefficients
  mbot = -4.4; // negative mass means no bottom running in the loop (consistent with small-x coefficients)

  // relative precision for integrals
  INTEGRAL_PRECISION = 15.e-3;

  // scheme at small x (set to "false" for Q0MSbar scheme)
  MSbar = true;

  // GLAP (collinear) improvement at the exponent (natural)
  // or at the level of coefficient function (unnatural, but easier and faster)
  APimprExponent = true;

  /***********************************************/



  // default values (modified by runtime arguments)
  double mH = 125;
  double sqrts = 8000;
  string set = "NNPDF23_nnlo_as_0118";

  // read runtime arguments
  read_arguments(argc, argv, mH, sqrts, set);




  // Start time counter
  struct timeval t0, t1;
  gettimeofday(&t0,NULL);




  // init
  double tau = mH*mH/sqrts/sqrts;
  sxc = new SmallxCoeffs("./src/MBDFV.dat");
  init_consts(mH);
  //
  cout << "Higgs mass:       " << mH << " GeV" << endl
       << "Collider energy:  " << sqrts << " GeV" << endl
       << "tau = mH^2/s =    " << tau << endl;
  //





  /******* RUN EXAMPLE *******/

  // set to false to show intermediate K-factor results
  quiet = false;
  //
  // init PDFs
  int subset = 0;
  pdf_init(set, subset, LHAPDF::SILENT);
  //
  //
  // Note also that the same PDF set has to be used at every order.
  double LO=1, NLO=0, NLOgg=0, NNLO=0, NNLOgg=0;
  //
  double LOpref = xs_LO_prefactor(mH);
  double muF = mH *_muF_mu_ratio;
  double muR = muF*_muR_muF_ratio;
  double alpha_s = alphas(muR);
  // results
  double rexact[2], rexact_gg[2], rsoft2[3], rsoft1[3], rNsoft[3], rhighenergy[3], momconsunc[3];

  // LO
  LO = tau*Lum(tau,muF)*LOpref*pow(alpha_s,2);

  // NLO
  order = 1;
  __all_channels__ = true;
  compute(mH, sqrts, LO, 0, 0, &rexact[0], NULL, NULL, NULL, NULL);
  NLO = LO + rexact[0];
  __all_channels__ = false;  // gg only
  compute(mH, sqrts, LO, 0, 0, &rexact_gg[0], &rsoft2[0], &rsoft1[0], &rNsoft[0], &rhighenergy[0], &momconsunc[0]);
  //compute(mH, sqrts, LO, 0, 0, &rexact_gg[0], NULL, NULL, NULL, NULL);
  NLOgg = LO + rexact_gg[0];

  // NNLO
  order = 2;
  __all_channels__ = true;
  compute(mH, sqrts, LO, NLO, 0, &rexact[1], NULL, NULL, NULL, NULL);
  NNLO = NLO + rexact[1];
  __all_channels__ = false;
  compute(mH, sqrts, LO, NLOgg, 0, &rexact_gg[1], &rsoft2[1], &rsoft1[1], &rNsoft[1], &rhighenergy[1], &momconsunc[1]);
  //compute(mH, sqrts, LO, NLOgg, 0, &rexact_gg[1], NULL, NULL, NULL, NULL);
  NNLOgg = NLOgg + rexact_gg[1];

  // NNNLO
  order = 3;
  compute(mH, sqrts, LO, NLOgg, NNLOgg, NULL, &rsoft2[2], &rsoft1[2], &rNsoft[2], &rhighenergy[2], &momconsunc[2]);

  // final results
  double appr1[3], appr2[3], appr_mean[3], ProcError[3];
  double herr = 0.05; // momentum violation (fraction)
  for(int i=0; i<3; i++) {
    appr1[i] = (rsoft1[i]+rhighenergy[i]);
    appr2[i] = (rsoft2[i]+rhighenergy[i]);
    appr_mean[i] = (appr1[i]+appr2[i])/2.;
    ProcError[i] = (fabs(appr_mean[i]-appr2[i])+herr*fabs(momconsunc[i]));  // estimated error for our procedure
  }
  //
  cout << endl
       << "The following cross-sections have been computed with the same PDF set (" << set << ") at every order."
       << endl << endl;
  cout << "LO:" << endl;
  cout << " exact     = " << LO << endl;
  cout << "NLO:" << endl;
  cout << " exact     = " << NLO << endl;
  cout << " exact(gg) = " << NLOgg << endl;
  cout << " approx    = " << LO + appr_mean[0] << " +/- " << ProcError[0] << endl;
  cout << "NNLO:" << endl;
  cout << " exact     = " << NNLO << endl;
  cout << " exact(gg) = " << NNLOgg << endl;
  cout << " approx    = " << NLO + appr_mean[1] << " +/- " << ProcError[1] << endl;
  cout << "NNNLO:" << endl;
  cout << " approx    = " << NNLO + appr_mean[2] << " +/- " << ProcError[2] << endl;
  //




  // Stop time counter
  gettimeofday(&t1,NULL);
  double t=t1.tv_sec-t0.tv_sec+(t1.tv_usec-t0.tv_usec)*0.000001;
  cout << endl << "Total time: " << t << "s" << endl;


  return 0;

}
